#include<stdio.h>
#include<stdlib.h>
#include "types.h"
#include "pin.h"
#include "defaults.h"
#include "mystring.h"
#include "warning.h"
#include "X.h"
#include "window.h"
#include "keyboard.h"
#include "save.h"
#include "input.h"
#include "edit.h"

static int gursor;    // vzdalenost od leveho okraje okenka ke kursoru
static int gelka;     // vzdalenost od leveho okraje okenka k pravemu konci textu
static int posunuti = 0;  // posunuti textu doleva, cili levu okraj textu = LOT-posunuti

static int cursor = 0; // ukazatel do textoveho pole
static int delka = 0;  // delka textoveho pole, textove pole nekonci standardni nulou

static char editpole[STRINGSIZE];   // aktualni radek, ktery se edituje
char *text_pole = editpole;

static int kursor_vyska;
static int font_y;

void loadfont()
{
  if ((font_info=XLoadQueryFont(display, FONTNAME)) == NULL)
    error("cannot open %s font", FONTNAME);
  
  kursor_vyska = font_info->ascent+font_info->descent;
  text_vyska = HOT+kursor_vyska+DOT;
  font_y = HOT+font_info->ascent;
  gursor = LOT;
  gelka = LOT;
}

void cleartext ()  // vycisti textove policko (ciste graficka zalezitost)
{
  XClearArea(display, textwin, 0, 0,
	     text_sirka, text_vyska, 0);
}

void redrawtext ()  // vykresli text, napr pri Expose.
{
  if(!delka) return;
  XDrawString(display, textwin, text_gc,
	      LOT-posunuti, font_y, text_pole, delka);

  // Nakreslim okraje po stranach:

  if(posunuti)
      XClearArea(display, textwin, 0, 0,
		 LOT, text_vyska, 0);

  if(gelka > text_sirka-POT)
      XClearArea(display, textwin, text_sirka-POT, 0,
		 POT, text_vyska, 0);
}

void ukazkursor ()  // vykresli kursor
{
  XFillRectangle(display, textwin, text_gc, gursor, HOT,
		 KURSIR, kursor_vyska);
}

void smazkursor ()  // smaze kursor
{
  XClearArea(display, textwin, gursor, HOT,
	     KURSIR, kursor_vyska, 0);

  if(cursor < delka)  // obnovim smazane pismeno
    XDrawString(display, textwin, text_gc, gursor, font_y,
		&(text_pole[cursor]), 1);
}

/*
Nasledujici funkce resi posouvani textu. To se ridi podle nasledujicich
pravidel (serazeno podle priority sestupne):
1) Levy okraj textu bude LOT pixelu od okraje okenka nebo vice vlevo.
   posunuti >= 0
2) Pravy okraj textu bude POT pixelu od okraje okenka nebo vice vpravo.
   gelka <= text_sirka-POT
3) Kursor bude alespon STRKAL pixelu od okraje.
   gursor >= STRKAL && gursor <= text_sirka-STRKAP
4) Text se posune co nejmene.
Kazde z techto pravidel musi byt splneno, neni-li to v konfliktu s nejakym
pravidlem s vyssi prioritou.
*/

int posuntext ()
{
  int ppos = 0;

  if(gursor < STRKAL) ppos = STRKAL-gursor;
  else if(gursor > text_sirka-STRKAP) ppos = text_sirka-STRKAP-gursor;

  if(gelka+ppos < text_sirka-POT) ppos = text_sirka-POT-gelka;

  if(posunuti-ppos < 0) ppos = posunuti;

  if(!ppos) return 0;

  posunuti -= ppos;
  gursor += ppos;
  gelka += ppos;
  cleartext();
  redrawtext();
  return 1;
}

void vyrestext ()  // vyresi zmenenou sirku textoveho okenka
{
  if(!okrajtext) return;
  posuntext();
  okraj_to_text();
  ukazkursor();
}

/*
Nasledujici kus kodu resi historii. Historie je ukladana do pole history, kde se
pro kazdy radek vymallokuje patricne dlouhy text. Ukladaji se tam standardni
C-ckovske stringy narozdil od pole editpole. Prvni radek se ulozi do history[0],
druhy do history[1] atd. Pote, co se historie ulozi do history[MAXHISTORY-1], je
ukladanaopet do history[0], history[1], ...
Posledni radek, do ktereho se ukladala historie udava promenna poslhist. Pokud je
historie prazdna, pak poslhist = -1. 
To, zda se uz historie zacala cyklit, cili pocet ulozenych radku dosahl maxima,
je ulozeno v promenne cyklhist.

Misto historie, ktere je zobrazovano je urceno podle promenne aktuhist. Pokud je
tato promenna mezi nulou a MAXHISTORY-1, jedna se o index pole history. Pokud
aktuhist==-1, je zobrazovan jeste nepotvrzeny radek a pri aktuhist==-2 je zobrazovan
prazdny radek.
*/

static char *history[MAXHISTORY];
static int poslhist = -1;
int aktuhist = -1;
static int cyklhist = 0;

static int ogursor;
static int ogelka;
static int oposunuti;
static int ocursor;
static int odelka;

/*
Funkce edithist zkopiruje zobrazovany radek historie do pole urcene k editaci textu.
Spucti se v okamziku, kdy se uzivatel pokusi neco pripsat ci smazat.
*/

void edithist ()
{
  int i;

  if(aktuhist == -1) return;  // Neni treba, pokud nejsem v historii.

  for(i=0; i<delka; i++) editpole[i] = text_pole[i]; // Zkopiruji string.

  aktuhist = -1;        // Vratim aktuhist na editpole.
  text_pole = editpole; // Vratim text_pole na editpole.
}

void ulozhist ()  // zkopiruje aktualni napis do historie
{
  int i;

  // Posunu poslhist.

  poslhist++;
  if(poslhist == MAXHISTORY){
    poslhist = 0;
    cyklhist = 1;
  }

  if(aktuhist == poslhist) return; /* Specialni pripad, kdy ukladam nejstarsi
				      historii a zaroven ji mazu */

  if(cyklhist) free(history[poslhist]);  // Uvolnim puvodni string.
  history[poslhist] = (char *)malloc((delka+1)*sizeof(char)); // Naalokuji novy string.
  for(i=0; i<delka; i++) history[poslhist][i] = text_pole[i]; // Zkopiruji string.
  history[poslhist][delka] = 0; // ZCeckuji ulozeny string.
}

/*
Funkce zalohujtext si ulozi informace o textu, takze editacni radek se pri vraceni
zobrazi stejne jako byl zobrazeny pri jeho opusteni.
*/

void zalohujtext (int *c, int *d, int *gc, int *gd, int *pos)
{
  *c = cursor;
  *d = delka;
  *gc = gursor;
  *gd = gelka;
  *pos = posunuti;
}

void odzaltext (int c, int d, int gc, int gd, int pos)
{
  cursor   = c;
  delka    = d;
  gursor   = gc;
  gelka    = gd;
  posunuti = pos;
}

/*
Funkce dohistory ukaze aktualni historii (z vety jdu DO HISTORie).
Pouziva se po zmene promenne aktuhist.
*/

void dohistory ()
{
  if(aktuhist == -2){  // Prazdny radek.
    cleartext();
    gursor = gelka = LOT;
    posunuti = cursor = delka = 0;
  }
  else if(aktuhist == -1){ // Editacni radek, zrestauruji ulozene hodnoty.
    cleartext();
    text_pole = editpole;
    odzaltext(ocursor, odelka, ogursor, ogelka, oposunuti);
    redrawtext();
  }
  else{  // Skutecna historie
    text_pole = history[aktuhist]; // Zmenim pointer text_pole.
    posunuti = 0;
    gelka = LOT;
    delka = strlen(history[aktuhist]); // Spocitam delku
    if(delka) gelka += XTextWidth(font_info, text_pole, delka); // Spocitam gelku

    // Kursor je na konci:
    cursor = delka;
    gursor = gelka;

    if(!posuntext()){ // Posunu text
      cleartext();    // nebo ho jen prekreslim.
      redrawtext();
    }
  }
  if(okrajtext) ukazkursor();
}

char pripisznak (char ch)
{
  int i, sirka;

  if(!okrajtext){
    okraj_to_text(); // Je-li aktivni okno, zaktivnim text.
    if(delka==STRINGSIZE){  // Doslo mi pole
      posuntext();
      ukazkursor();
      return 0;
    }
  }
  else if(delka==STRINGSIZE) return 0; // Doslo mi pole

  edithist();  // Prepnu se do editacniho radku. (pokud tam nejsem)

  for(i=delka-1; i >= cursor; i--)
    text_pole[i+1] = text_pole[i];  // Posunu znaku za kursorem. (v poli)
  text_pole[cursor] = ch; // Ulozim novy znak

  sirka = XTextWidth(font_info, &ch, 1); // Zjistim sirku pismena.

  // Posunu kursor, zvetsila se delka:
  delka++;
  cursor++;
  gursor += sirka;
  gelka += sirka;

  if(posuntext()){ // Pokud se mi text posune, mam grafiku vyresenou.
    ukazkursor();
    if(cursor == delka) return -1;
    else return 1;
  }

  XClearArea(display, textwin, gursor-sirka, 0, // Smazu kursor nebo text za kursorem.
	     (gelka-gursor > KURSIR) ? (gelka-gursor):KURSIR,
	     text_vyska, 0);

  // Vykreslim text, co jsem napsal a posunul:
  XDrawString(display, textwin, text_gc, gursor-sirka, font_y, 
	      &(text_pole[cursor-1]), delka-cursor+1);

  if(gelka > text_sirka-POT) // cerny okraj na prave strane textoveho okenka
    XClearArea(display, textwin, text_sirka-POT, 0,
	       POT, text_vyska, 0);
  ukazkursor();
  if(cursor == delka) return -1;
  else return 1;
}

/*
Nasledujici funkce jsou provazany s urcitou klavesou. Tato vazba je vytvorena
v bindings.c
*/

void prepniokraj (int natext) // Vazba s klavesou Tab.
{
  if(natext){ // Da okraj na textwin
    okrajtext = 1;
    vyrestext();
  }
  else{ // Da okraj na win
    smazkursor();
    okraj_to_win();
  }
}

char delete (int backspace) // Backspace, pokud parametr backspace je true. Jinak Delete.
{
  int i;

  if(backspace){   // Prevedu backspace na delete
    if(!cursor) return 0;
    cursor--;
    gursor -= XTextWidth(font_info, &(text_pole[cursor]), 1);
  }

  if(delka == cursor) return 0; // Jsem na konci radku.

  edithist();  // Prepnu se do editacniho radku. (pokud tam nejsem)

  // Smazu text pred kursorem i s kursorem, co tam mohl zbyt z backspace.
  XClearArea(display, textwin, gursor, 0,
	     gelka-gursor+KURSIR, text_vyska, 0);

  delka--;    // Zmensila se delka textu.
  gelka -= XTextWidth(font_info, &(text_pole[cursor]), 1);

  for(i=cursor; i<delka+1; i++) // posunu zbyle znaky za kursorem.
    text_pole[i] = text_pole[i+1];

  if(posuntext()){ // Pokud se mi text posune, mam grafiku vyresenou.
    ukazkursor();
    return 1;
  }
  if(delka != cursor)
    XDrawString(display, textwin, text_gc, gursor,  // Vykreslim zbyle znaky za kursorem.
		font_y, &(text_pole[cursor]), delka-cursor);

  ukazkursor();
  return 1;
}

void posunkursor (int jak)
{
  // jak: 0=left, 1=right, 2=home, 3=end

  if((!(jak%2) && !cursor) || ((jak%2) && cursor == delka)) return;
     // Nemohu posunou kursor podle promenne jak.

  smazkursor();

  // Zmenim cursor a gursor:

  switch((jak)){
  case 0:
    cursor--;
    gursor -= XTextWidth(font_info, &(text_pole[cursor]), 1);
    break;
  case 1:
    gursor += XTextWidth(font_info, &(text_pole[cursor]), 1);
    cursor++;
    break;
  case 2:
    cursor = 0;
    gursor = LOT-posunuti;
    break;
  case 3:
    gursor = gelka;
    cursor = delka;
  }
  posuntext();  // Pripadne posunu text.

  ukazkursor();
}

void zkoumejhist (int jak)
{
  // jak: 0=down, 1=up, 2=pgdown, 3=pgup

  if((!(jak%2) && (aktuhist == -2 || (aktuhist == -1 && !delka))) ||
     ((jak%2) && (aktuhist == (poslhist+1)%MAXHISTORY ||
		     (aktuhist == 0 && !cyklhist) ||
		     (poslhist == -1 && aktuhist == -1)))) return;
     // Nemohu jit do historie podle promenne jak.
  
  if(aktuhist == -1) zalohujtext(&ocursor, &odelka, &ogursor, &ogelka, &oposunuti);
         // Jsem-li na editacnim radku, ulozim si to.

  // Posunu aktuhist:

  switch(jak){
  case 0:
    if(aktuhist == -1) aktuhist = -2;
    else if(aktuhist == poslhist) aktuhist = -1;
    else{
      aktuhist++;
      if(aktuhist == MAXHISTORY) aktuhist = 0;
    }
    break;
  case 1:
    if(aktuhist == -1) aktuhist = poslhist;
    else if(aktuhist == -2) aktuhist = -1;
    else{
      aktuhist--;
      if(aktuhist < 0) aktuhist += MAXHISTORY;
    }
    break;
  case 2:
    if(aktuhist == -1 || odelka) aktuhist = -2;
    else aktuhist = -1;
    break;
  case 3:
    if(poslhist == -1) aktuhist = -1;
    else if(!cyklhist || poslhist == MAXHISTORY-1) aktuhist = 0;
    else aktuhist = poslhist+1;
  }

  dohistory(); // Aplikuji aktuhist.
  if(!okrajtext) mengraf = 1;
}

void text_enter ()
{
  char povedlo;

  if(okrajtext) smazkursor();

  if(delka != 0){
    ulozhist(); // Ulozim text do historie,
    aktuhist = -1;
    text_pole = editpole;

    // smazu ho:

    cleartext();
    posunuti = cursor = delka = 0;
    gursor = gelka = LOT;

    // a provedu

    povedlo = provedstr(history[poslhist]);
  }


  if(okrajtext){
    if(povedlo) okraj_to_win(); // Nakonec prepnu zpatky do chozeni.
    else ukazkursor();
  }

  return;
}

void text_escape ()
{
  int orihist;

  okraj_to_win();
  orihist = aktuhist;
  zkoumejhist(2);
  if(aktuhist == orihist) smazkursor();
  return;
}

/*******************************************************************/
/*********************** FUNKCE PRO MODUL GSL **********************/
/*******************************************************************/

void zmentext(char *str)
{
  for(delka=0; str[delka] && delka < STRINGSIZE; delka++) text_pole[delka] = str[delka];
  if(delka == STRINGSIZE && str[STRINGSIZE])
    warning("zmentext: prilis dlouhy text:\n %s\n", str);
  posunuti = 0;
  gelka = LOT;
  if(delka) gelka += XTextWidth(font_info, text_pole, delka); // Spocitam gelku

  // Kursor je na konci:
  cursor = delka;
  gursor = gelka;

  if(!posuntext()){ // Posunu text
    cleartext();    // nebo ho jen prekreslim.
    redrawtext();
  }
  if(okrajtext) ukazkursor();
}

char *textcstr()
{
  if(delka == STRINGSIZE) return dupnstr(text_pole, delka);
  text_pole[delka] = 0;
  return text_pole;
}


/*******************************************************************/
/****************** FUNKCE PRO MODULY SAVE A LOAD ******************/
/*******************************************************************/

void save_hist ()
{
  int i, j;

  if(cyklhist) save_unsint(MAXHISTORY);
  else save_unsint(poslhist+1);

  save_unsint(delka);
  for(i = poslhist;;){
    if(!cyklhist && i < 0) break;
    save_unsint(strlen(history[i]));
    i--;
    if(i == poslhist) break;
  }
  save_cs();
  for(j=0; j<delka; j++) save_char(editpole[j]);
  for(i = poslhist;;){
    if(!cyklhist && i < 0) break;
    for(j=0; history[i][j]; j++) save_char(history[i][j]);
    i--;
    if(i == poslhist) break;
  }
}

void apply_loaded_hist (char **hist, int poceth)
{
  int i;

  poslhist = poceth-1;
  delka = strlen(hist[poceth]);
  for(i = 0; hist[poceth][i]; i++) editpole[i] = hist[poceth][i];
  free(hist[poceth]);
  for(i = 0; i < poceth; i++) history[i] = hist[i];
}
