#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include "types.h"
#include "defaults.h"
#include "warning.h"
#include "mystring.h"
#include "color.h"
#include "filer.h"
#include "map.h"

kq_map *mapa = NULL;
float **mapa_z = NULL;
kq_submap *prvsubmap = NULL;
kq_submap *bgmap = NULL;
kq_submap *prvhidsm = NULL;
char autoreloadmap = 0;

static kq_color ***allocpole (int sirka, int vyska)
/*************************************************/
{
  kq_color ***vysledek;
  int x;

  vysledek = (kq_color ***)malloc(sizeof(kq_color **)*sirka);
  for(x=0; x<sirka; x++)
    vysledek[x] = (kq_color **)malloc(sizeof(kq_color *)*vyska);

  return vysledek;
}

kq_map nactimapu (char *filename)
/*******************************/
{
  kq_filer *soubor;
  int znak, ch, i, j, k, pocetbarev;
  kq_color *pom, *pomo;
  kq_map vysledek;
  char *fullname;

  for(pom = prvbarva; pom != NULL; pom = pom->next) pom->znak = 0;

  fullname = joinstrings(DATADIR, MAPDIR, filename, MAPSUFFIX, NULL);

  soubor = mf_otevri(fullname);
  mf_musibyt(soubor, "/* XPM */");
  mf_najdi(soubor, '"');

  vysledek.sirka = mf_cislo(soubor);
  mf_musibyt(soubor, " ");
  vysledek.vyska = mf_cislo(soubor);
  mf_musibyt(soubor, " ");
  pocetbarev = mf_cislo(soubor);
  mf_musibyt(soubor, " 1\"");

  for(i=0; i<pocetbarev; i++){
    mf_najdi(soubor, '"');
    znak=mf_znak(soubor);
    mf_musibyt(soubor, "	c ");

    ch=1;
    pom = prvbarva;
    for(j=0; ch; j++){
      ch = mf_znak(soubor);
      if(ch == EOF){
	warning ("Konec souboru pri nacitani %d. barvy", j);
	mf_error(soubor);
      }
      if(ch == '"'){
	ch=0;
	soubor->ch = '"';
      }
      for(pomo = pom; pom; pom = pom->next){
	for(k=0; k<j && pom->nazev[k] == pomo->nazev[k]; k++);
	if(k==j && pom->nazev[k] == ch) break;
      }
      if(pom == NULL) break;
    }
    if(pom != NULL) pom->znak = znak;
    mf_najdi(soubor, '"');
  }

  vysledek.pole = allocpole(vysledek.sirka, vysledek.vyska);

  for(i=0; i<vysledek.vyska; i++){
    mf_najdi(soubor, '"');
    for(j=0; j<vysledek.sirka; j++){
      znak = mf_znak(soubor);
      for(pom=prvbarva; pom; pom=pom->next)
	if(znak && znak == pom->znak){
	  vysledek.pole[j][i] = pom;
	  break;
	}
      if(pom == NULL){
	warning ("Neznama barva: %d='%c'", znak, znak);
	mf_error(soubor);
      }
    }
    mf_musibyt(soubor, "\"");
  }
  mf_zavri(soubor);
  free(fullname);

  return vysledek;
}

static void freemap (kq_map map)
/****************************/
{
  int i;

  for(i=0; i < map.sirka; i++) free(map.pole[i]);
  free(map.pole);
}

static kq_submap *nejvyssipix (int x, int y)
{
  kq_submap *pom, *max;
  kq_color *c;
  float z;

  max = NULL;
  for(pom = prvsubmap; pom; pom = pom->next){
    c = smpixel(pom, x,y);
    if(c && (max == NULL || c->z+pom->z > z)){
      z = c->z+pom->z;
      max = pom;
    }
  }

  return max;
}

static void pouzijsubmap (kq_submap *sm)
{
  int x, y;

  if(!mapa) return;

  for(x = (sm->x>0 ? sm->x:0);
      x < sm->x+sm->map.sirka && x < mapa->sirka; x++)
    for(y = (sm->y>0 ? sm->y:0);
	y < sm->y+sm->map.vyska && y < mapa->vyska; y++)
      if(sm->z + sm->map.pole[x-sm->x][y-sm->y]->z >= mapa_z[x][y]){
	mapa_z[x][y] = sm->z + sm->map.pole[x-sm->x][y-sm->y]->z;
	mapa->pole[x][y] = sm->map.pole[x-sm->x][y-sm->y];
      }
}

static void zrussubmap (kq_submap *sm)
{
  int x, y;
  kq_submap *pom;

  for(x = (sm->x>0 ? sm->x:0);
      x < sm->x+sm->map.sirka && x < mapa->sirka; x++)
    for(y = (sm->y>0 ? sm->y:0);
	y < sm->y+sm->map.vyska && y < mapa->vyska; y++)
      if(mapa->pole[x][y] == sm->map.pole[x-sm->x][y-sm->y]){
	pom = nejvyssipix(x, y);
	mapa->pole[x][y] = smpixel(pom, x, y);
	mapa_z[x][y] = mapa->pole[x][y]->z + sm->z;
      }
}

static void zaradsubmap (kq_submap *sm, kq_submap **prvni)
{
  while(*prvni && (*prvni)->z > sm->z) prvni = &(*prvni)->next;

  sm->next = *prvni;
  *prvni = sm;
}

static void vyradsubmap (kq_submap *sm, kq_submap **prvni)
{
  while(*prvni && *prvni != sm) prvni = &(*prvni)->next;

  *prvni = sm->next;
}

void zmenbgmap (char *filename)
{
  if(bgmap){
    freemap(bgmap->map);
    free(bgmap->nazev);
  }

  if(!filename){
    if(bgmap){
      vyradsubmap(bgmap, &prvsubmap);
      free(bgmap);
      bgmap = NULL;
    }
    return;
  }

  if(!bgmap){
    bgmap = (kq_submap *)malloc(sizeof(kq_submap));
    bgmap->x = 0;
    bgmap->y = 0;
    bgmap->z = 0;
    bgmap->hide = 0;
    bgmap->lokalni = 0;
    zaradsubmap(bgmap, &prvsubmap);
  }

  bgmap->map = nactimapu(filename);
  bgmap->nazev = strdup(filename);
  if(autoreloadmap) reloadmap();
}

kq_submap *vytvorsubmap (int x, int y, float z, char *filename,
		       char lokalni, char hide)
{
  kq_submap *vysledek;

  vysledek = (kq_submap *)malloc(sizeof(kq_submap));;
  vysledek->x = x;
  vysledek->y = y;
  vysledek->z = z;
  vysledek->hide = hide;
  vysledek->lokalni = lokalni;
  vysledek->map = nactimapu(filename);
  vysledek->nazev = strdup(filename);

  if(hide) zaradsubmap(vysledek, &prvhidsm);
  else{
    zaradsubmap(vysledek, &prvsubmap);
    if(autoreloadmap) pouzijsubmap(vysledek);
  }

  return vysledek;
}

void reloadmap ()
{
  int x,y;
  kq_submap *pom;

  if(!bgmap){
    if(mapa){
      for(x=0; x<mapa->sirka; x++) free(mapa_z[x]);
      free(mapa_z);
      freemap(*mapa);
      free(mapa);
      mapa = NULL;
    }
    return;
  }

  if(!mapa ||
     mapa->sirka != bgmap->map.sirka ||
     mapa->vyska != bgmap->map.vyska){

    if(mapa){
      for(x=0; x<mapa->sirka; x++) free(mapa_z[x]);
      free(mapa_z);
      if(mapa) freemap(*mapa);
    }
    else mapa = (kq_map *)malloc(sizeof(kq_map));

    mapa->sirka = bgmap->map.sirka;
    mapa->vyska = bgmap->map.vyska;
    mapa->pole = allocpole(mapa->sirka, mapa->vyska);

    mapa_z = (float **)malloc(sizeof(float *)*mapa->sirka);
    for(x=0; x<mapa->sirka; x++)
      mapa_z[x] = (float *)malloc(sizeof(float)*mapa->vyska);
  }

  for(x=0; x<mapa->sirka; x++)
    for(y=0; y<mapa->vyska; y++){
      mapa->pole[x][y] = bgmap->map.pole[x][y];
      mapa_z[x][y] = bgmap->map.pole[x][y]->z;
    }

  for(pom = prvsubmap; pom; pom = pom->next)
    if(pom != bgmap) pouzijsubmap(pom);
}

kq_color *mappixel (kq_map *map, int x, int y)
{
  if(x < 0 || y < 0 || x >= map->sirka || y >= map->vyska) return NULL;
  return map->pole[x][y];
}

kq_color *smpixel (kq_submap *sm, int x, int y)
{
  return mappixel(&sm->map, x-sm->x, y-sm->y);
}

kq_submap *najdibarvu (kq_color *col, int x, int y)
{
  kq_submap *pom;

  for(pom = prvsubmap; pom; pom = pom->next)
    if(col == smpixel(pom, x,y)) return pom;

  return NULL;
}

void hidesubmap (kq_submap *sm)
{
  if(sm == bgmap){
    warning("hidesubmap(bgmap)");
    return;
  }

  if(sm->hide) return;

  vyradsubmap(sm, &prvsubmap);
  if(autoreloadmap) zrussubmap(sm);

  sm->hide = 1;

  sm->next = prvhidsm;
  prvhidsm = sm;
}

void unhidesubmap (kq_submap *sm)
{
  if(sm == bgmap){
    warning("unhidesubmap(bgmap)");
    return;
  }

  if(!sm->hide) return;

  vyradsubmap(sm, &prvhidsm);
  
  sm->hide = 0;

  zaradsubmap(sm, &prvsubmap);
  if(autoreloadmap) pouzijsubmap(sm);
}

static void freesm (kq_submap *sm)
{
  freemap(sm->map);
  free(sm->nazev);
  free(sm);
}

void znicsubmap (kq_submap *sm)
{
  if(sm == bgmap){
    warning("znicsubmap(bgmap)");
    return;
  }

  if(sm->hide) vyradsubmap(sm, &prvhidsm);
  else{
    vyradsubmap(sm, &prvsubmap);
    if(autoreloadmap) zrussubmap(sm);
  }

  freesm(sm);
}

void deletelsubmaps ()
{
  kq_submap **pom, *pomo, *pomoc;
  char trebarel;

  if(bgmap -> lokalni){
    warning("deletelsubmaps: lokalni bgmap");
    bgmap->lokalni = 0;
  }

  for(pom = &prvhidsm;; pom = &pomo->next){
    for(pomo = *pom; pomo && pomo->lokalni > 0; pomo = pomoc){
      pomoc = pomo->next;
      freesm(pomo);
    }
    *pom = pomo;
    if(!pomo) break;
  }

  trebarel = 0;
  for(pom = &prvsubmap;; pom = &pomo->next){
    for(pomo = *pom; pomo && pomo->lokalni; pomo = pomoc){
      pomoc = pomo->next;

      trebarel = 1;

      if(pomo->lokalni > 0) freesm(pomo);
      else{
	pomo->next = prvhidsm;
	prvhidsm = pomo;
      }
    }
    *pom = pomo;
    if(!pomo) break;
  }

  if(autoreloadmap && trebarel) reloadmap();
}

void deleteAsubmaps ()
{
  kq_submap *pom;
  int x;

  while(prvsubmap){
    pom = prvsubmap->next;
    freesm(prvsubmap);
    prvsubmap = pom;
  }
  while(prvhidsm){
    pom = prvhidsm->next;
    freesm(prvhidsm);
    prvhidsm = pom;
  }

  bgmap = NULL;
  if(autoreloadmap){
    for(x=0; x<mapa->sirka; x++) free(mapa_z[x]);
    free(mapa_z);
    mapa = NULL;
  }
}
